@@ -14,12 +14,14 @@ from typing import Dict
 from typing import Iterable
 from typing import Iterator
 from typing import Optional
+from typing import Set
 from typing import Tuple
 from typing import Type
 from typing import Union
 
 import sqlparse  # type: ignore
-from tortoise import Model
+from tortoise import ForeignKeyFieldInstance
+from tortoise import Model as TortoiseModel
 from tortoise import ModuleType
 from tortoise import Tortoise
 from tortoise import connections
@@ -27,7 +29,6 @@ from tortoise.backends.asyncpg.client import AsyncpgDBClient
 from tortoise.backends.base.client import BaseDBAsyncClient
 from tortoise.backends.sqlite.client import SqliteClient
 from tortoise.fields import DecimalField
-from tortoise.transactions import in_transaction
 from tortoise.utils import get_schema_sql
 
 from dipdup.exceptions import DatabaseConfigurationError
@@ -44,7 +45,7 @@ def get_connection() -> BaseDBAsyncClient:
     return connections.get(DEFAULT_CONNECTION_NAME)
 
 
-def _set_connection(conn: BaseDBAsyncClient) -> None:
+def set_connection(conn: BaseDBAsyncClient) -> None:
     connections.set(DEFAULT_CONNECTION_NAME, conn)
 
 
@@ -70,9 +71,9 @@ async def tortoise_wrapper(url: str, models: Optional[str] = None, timeout: int
                     modules=model_modules,
                 )
 
-                # FIXME: Wait for the connection to be ready, required since 0.19.0
                 conn = get_connection()
                 await conn.execute_query('SELECT 1')
+            # FIXME: Logging
             except OSError:
                 _logger.warning('Can\'t establish database connection, attempt %s/%s', attempt, timeout)
                 if attempt == timeout - 1:
@@ -85,24 +86,14 @@ async def tortoise_wrapper(url: str, models: Optional[str] = None, timeout: int
         await Tortoise.close_connections()
 
 
-@asynccontextmanager
-async def in_global_transaction():
-    """Enforce using transaction for all queries inside wrapped block. Works for a single DB only."""
-    try:
-        original_conn = get_connection()
-        async with in_transaction() as conn:
-            _set_connection(conn)
-            yield
-    finally:
-        _set_connection(original_conn)
-
-
 def is_model_class(obj: Any) -> bool:
     """Is subclass of tortoise.Model, but not the base class"""
-    return isinstance(obj, type) and issubclass(obj, Model) and obj != Model and not getattr(obj.Meta, 'abstract', False)
+    from dipdup.models import Model
+
+    return isinstance(obj, type) and issubclass(obj, TortoiseModel) and obj not in (TortoiseModel, Model)
 
 
-def iter_models(package: Optional[str]) -> Iterator[Tuple[str, Type[Model]]]:
+def iter_models(package: Optional[str]) -> Iterator[Tuple[str, Type[TortoiseModel]]]:
     """Iterate over built-in and project's models"""
     modules = [
         ('int_models', importlib.import_module('dipdup.models')),
@@ -125,23 +116,6 @@ def iter_models(package: Optional[str]) -> Iterator[Tuple[str, Type[Model]]]:
                 yield app, attr_value
 
 
-def set_decimal_context(package: str) -> None:
-    """Adjust system decimal context to match database precision"""
-    context = decimal.getcontext()
-    prec = context.prec
-    for _, model in iter_models(package):
-        for field in model._meta.fields_map.values():
-            if isinstance(field, DecimalField):
-                prec = max(prec, field.max_digits)
-
-    if context.prec < prec:
-        _logger.warning('Decimal context precision has been updated: %s -> %s', context.prec, prec)
-        context.prec = prec
-        # NOTE: DefaultContext used for new threads
-        decimal.DefaultContext.prec = prec
-        decimal.setcontext(context)
-
-
 def get_schema_hash(conn: BaseDBAsyncClient) -> str:
     """Get hash of the current schema"""
     schema_sql = get_schema_sql(conn, False)
@@ -191,22 +165,29 @@ async def truncate_schema(conn: BaseDBAsyncClient, name: str) -> None:
     await conn.execute_script(f"SELECT truncate_schema('{name}')")
 
 
-async def wipe_schema(conn: BaseDBAsyncClient, name: str, immune_tables: Tuple[str, ...]) -> None:
+async def wipe_schema(
+    conn: BaseDBAsyncClient,
+    schema_name: str,
+    immune_tables: Set[str],
+) -> None:
+    """Truncate schema preserving immune tables. Executes in a transaction"""
     if isinstance(conn, SqliteClient):
         raise NotImplementedError
 
-    immune_schema_name = f'{name}_immune'
-    if immune_tables:
-        await create_schema(conn, immune_schema_name)
-        for table in immune_tables:
-            await move_table(conn, table, name, immune_schema_name)
+    immune_schema_name = f'{schema_name}_immune'
 
-    await truncate_schema(conn, name)
+    async with conn._in_transaction() as conn:
+        if immune_tables:
+            await create_schema(conn, immune_schema_name)
+            for table in immune_tables:
+                await move_table(conn, table, schema_name, immune_schema_name)
 
-    if immune_tables:
-        for table in immune_tables:
-            await move_table(conn, table, immune_schema_name, name)
-        await drop_schema(conn, immune_schema_name)
+        await truncate_schema(conn, schema_name)
+
+        if immune_tables:
+            for table in immune_tables:
+                await move_table(conn, table, immune_schema_name, schema_name)
+            await drop_schema(conn, immune_schema_name)
 
 
 async def drop_schema(conn: BaseDBAsyncClient, name: str) -> None:
@@ -226,34 +207,70 @@ async def move_table(conn: BaseDBAsyncClient, name: str, schema: str, new_schema
 
 def prepare_models(package: Optional[str]) -> None:
     """Prepare TortoiseORM models to use with DipDup.
-    Generate missing table names, validate models, increase decimal precision.
+    Generate missing table names, validate models, increase decimal precision if needed.
     """
-    from dipdup.models import Model
+    # NOTE: Circular imports
+    import dipdup.models
 
+    db_tables: Set[str] = set()
     decimal_context = decimal.getcontext()
     prec = decimal_context.prec
 
     for app, model in iter_models(package):
-
         # NOTE: Enforce our class for user models
-        if app == 'models' and not issubclass(model, Model):
-            raise DatabaseConfigurationError('Project models must be subclassed from `dipdup.models.Model`', model)
+        if app == 'models' and not issubclass(model, dipdup.models.Model):
+            raise DatabaseConfigurationError(
+                'Project models must be subclassed from `dipdup.models.Model`.'
+                '\n\n'
+                'Replace `from tortoise import Model` import with `from dipdup.models import Model`.',
+                model,
+            )
 
         # NOTE: Generate missing table names before Tortoise does
         if not model._meta.db_table:
             model._meta.db_table = pascal_to_snake(model.__name__)
 
+        if model._meta.db_table not in db_tables:
+            db_tables.add(model._meta.db_table)
+        else:
+            raise DatabaseConfigurationError(
+                'Table name is duplicated or reserved. Make sure that all models have unique table names.',
+                model,
+            )
+
         # NOTE: Enforce tables in snake_case
         table_name = model._meta.db_table
         if table_name != pascal_to_snake(table_name):
-            raise DatabaseConfigurationError('Table name must be in snake_case', model)
+            raise DatabaseConfigurationError(
+                'Table name must be in snake_case.',
+                model,
+            )
 
         for field in model._meta.fields_map.values():
             # NOTE: Enforce fields in snake_case
             field_name = field.model_field_name
-
             if field_name != pascal_to_snake(field_name):
-                raise DatabaseConfigurationError('Model fields must be in snake_case', model)
+                raise DatabaseConfigurationError(
+                    'Model fields must be in snake_case.',
+                    model,
+                    field_name,
+                )
+
+            # NOTE: Enforce unique field names to avoid GraphQL issues
+            if field_name == table_name:
+                raise DatabaseConfigurationError(
+                    'Model field names must differ from table name.',
+                    model,
+                    field_name,
+                )
+
+            # NOTE: The same for backward relations
+            if isinstance(field, ForeignKeyFieldInstance) and field.related_name == table_name:
+                raise DatabaseConfigurationError(
+                    'Model field names must differ from table name.',
+                    model,
+                    f'related_name={field.related_name}',
+                )
 
             # NOTE: Increase decimal precision if needed
             if isinstance(field, DecimalField):
